@@ -3,7 +3,8 @@
 
    (C) 2002 Rusty Russell IBM Corporation
  */
-#include <stdarg.h>
+#define _GNU_SOURCE /* asprintf */
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <getopt.h>
@@ -22,6 +23,8 @@
 
 #include "zlibsupport.h"
 #include "depmod.h"
+#include "logging.h"
+#include "index.h"
 #include "moduleops.h"
 #include "tables.h"
 
@@ -56,39 +59,6 @@ struct module_search
 static int verbose;
 static unsigned int skipchars;
 
-void fatal(const char *fmt, ...)
-{
-	va_list arglist;
-
-	fprintf(stderr, "FATAL: ");
-
-	va_start(arglist, fmt);
-	vfprintf(stderr, fmt, arglist);
-	va_end(arglist);
-
-	exit(1);
-}
-
-void warn(const char *fmt, ...)
-{
-	va_list arglist;
-
-	fprintf(stderr, "WARNING: ");
-
-	va_start(arglist, fmt);
-	vfprintf(stderr, fmt, arglist);
-	va_end(arglist);
-}
-
-void *do_nofail(void *ptr, const char *file, int line, const char *expr)
-{
-	if (!ptr) {
-		fatal("Memory allocation failure %s line %d: %s.\n",
-		      file, line, expr);
-	}
-	return ptr;
-}
-
 #define SYMBOL_HASH_SIZE 1024
 struct symbol
 {
@@ -209,6 +179,7 @@ static struct option options[] = { { "all", 0, NULL, 'a' },
 				   { "verbose", 0, NULL, 'v' },
 				   { "version", 0, NULL, 'V' },
 				   { "config", 1, NULL, 'C' },
+				   { "warn", 1, NULL, 'w' },
 				   { NULL, 0, NULL, 0 } };
 
 /* Version number or module name?  Don't assume extension. */
@@ -275,9 +246,9 @@ static void print_usage(const char *name)
 {
 	fprintf(stderr,
 	"%s " VERSION " -- part of " PACKAGE "\n"
-	"%s -[aA] [-n -e -v -q -V -r -u]\n"
+	"%s -[aA] [-n -e -v -q -V -r -u -w]\n"
 	"      [-b basedirectory] [forced_version]\n"
-	"depmod [-n -e -v -q -r -u] [-F kernelsyms] module1.ko module2.ko ...\n"
+	"depmod [-n -e -v -q -r -u -w] [-F kernelsyms] module1.ko module2.ko ...\n"
 	"If no arguments (except options) are given, \"depmod -a\" is assumed\n"
 	"\n"
 	"depmod will output a dependancy list suitable for the modprobe utility.\n"
@@ -290,6 +261,7 @@ static void print_usage(const char *name)
 	"\t-e, --errsyms        Report not supplied symbols\n"
 	"\t-V, --version        Print the release version\n"
 	"\t-v, --verbose        Enable verbose mode\n"
+	"\t-w, --warn		Warn on duplicates\n"
 	"\t-h, --help           Print this usage message\n"
 	"\n"
 	"The following options are useful for people managing distributions:\n"
@@ -328,6 +300,13 @@ int needconv(const char *elfhdr)
 		abort();
 }
 
+static char *my_basename(const char *name)
+{
+	const char *base = strrchr(name, '/');
+	if (base) return (char *)base + 1;
+	return (char *)name;
+}
+
 static struct module *grab_module(const char *dirname, const char *filename)
 {
 	struct module *new;
@@ -338,6 +317,7 @@ static struct module *grab_module(const char *dirname, const char *filename)
 		sprintf(new->pathname, "%s/%s", dirname, filename);
 	else
 		strcpy(new->pathname, filename);
+	new->basename = my_basename(new->pathname);
 
 	INIT_LIST_HEAD(&new->dep_list);
 
@@ -393,13 +373,6 @@ static int in_loop(struct module *mod, const struct module_traverse *traverse)
 	return 0;
 }
 
-static char *basename(const char *name)
-{
-	const char *base = strrchr(name, '/');
-	if (base) return (char *)base + 1;
-	return (char *)name;
-}
-
 /* Assume we are doing all the modules, so only report each loop once. */
 static void report_loop(const struct module *mod,
 			const struct module_traverse *traverse)
@@ -421,8 +394,8 @@ static void report_loop(const struct module *mod,
 
 	warn("Loop detected: %s ", mod->pathname);
 	for (i = traverse->prev; i->prev; i = i->prev)
-		fprintf(stderr, "needs %s ", basename(i->mod->pathname));
-	fprintf(stderr, "which needs %s again!\n", basename(mod->pathname));
+		fprintf(stderr, "needs %s ", i->mod->basename);
+	fprintf(stderr, "which needs %s again!\n", i->mod->basename);
 }
 
 /* This is damn slow, but loops actually happen, and we don't want to
@@ -458,14 +431,45 @@ static void order_dep_list(struct module *start, struct module *mod)
 	}
 }
 
+static struct module *deleted = NULL;
+
 static void del_module(struct module **modules, struct module *delme)
 {
 	struct module **i;
 
 	/* Find pointer to it. */ 
-	for (i = modules; *i != delme; i = &(*i)->next);
+	if (modules) {
+		for (i = modules; *i != delme; i = &(*i)->next);
+		
+		*i = delme->next;
+	}
+	
+	/* Save on a list to quiet valgrind.
+	   Can't free - other modules may depend on them */
+	delme->next = deleted;
+	deleted = delme;
+}
 
-	*i = delme->next;
+/* Convert filename to the module name.  Works if filename == modname, too. */
+static void filename2modname(char *modname, const char *filename)
+{
+	const char *afterslash;
+	unsigned int i;
+
+	afterslash = strrchr(filename, '/');
+	if (!afterslash)
+		afterslash = filename;
+	else
+		afterslash++;
+
+	/* Convert to underscores, stop at first . */
+	for (i = 0; afterslash[i] && afterslash[i] != '.'; i++) {
+		if (afterslash[i] == '-')
+			modname[i] = '_';
+		else
+			modname[i] = afterslash[i];
+	}
+	modname[i] = '\0';
 }
 
 static void output_deps(struct module *modules,
@@ -473,21 +477,6 @@ static void output_deps(struct module *modules,
 {
 	struct module *i;
 
-	for (i = modules; i; i = i->next)
-		i->ops->calculate_deps(i, verbose);
-
-	/* Strip out loops. */
- again:
-	for (i = modules; i; i = i->next) {
-		if (has_dep_loop(i, NULL)) {
-			warn("Module %s ignored, due to loop\n",
-			     i->pathname + skipchars);
-			del_module(&modules, i);
-			goto again;
-		}
-	}
-
-	/* Now dump them out. */
 	for (i = modules; i; i = i->next) {
 		struct list_head *j, *tmp;
 		order_dep_list(i, i);
@@ -503,6 +492,46 @@ static void output_deps(struct module *modules,
 	}
 }
 
+/* warn whenever duplicate module aliases, deps, or symbols are found. */
+int warn_dups = 0;
+
+static void output_deps_bin(struct module *modules,
+			FILE *out)
+{
+	struct module *i;
+	struct index_node *index;
+	char *line;
+	char *p;
+
+	index = index_create();
+
+	for (i = modules; i; i = i->next) {
+		struct list_head *j, *tmp;
+		char modname[strlen(i->pathname)+1];
+		
+		order_dep_list(i, i);
+		
+		filename2modname(modname, i->pathname + skipchars);
+		asprintf(&line, "%s %s:", modname, i->pathname + skipchars);
+		p = line;
+		list_for_each_safe(j, tmp, &i->dep_list) {
+			struct module *dep
+				= list_entry(j, struct module, dep_list);
+			asprintf(&line, "%s %s", p, dep->pathname + skipchars);
+			free(p);
+			p = line;
+			list_del_init(j);
+		}
+		if (index_insert(index, line) && warn_dups)
+			warn("duplicate module deps:\n%s\n",line);
+		free(line);
+	}
+	
+	index_write(index, out);
+	index_destroy(index);
+}
+
+
 static int smells_like_module(const char *name)
 {
 	return ends_in(name,".ko") || ends_in(name, ".ko.gz");
@@ -601,7 +630,7 @@ static struct module *do_module(const char *dirname,
 	/* Check if module is already in the list. */
 	for (i = &list; *i; i = &(*i)->next) {
 
-		if (streq(basename((*i)->pathname), filename)) {
+		if (streq((*i)->basename, filename)) {
 			char newpath[strlen(dirname) + strlen("/")
 				      + strlen(filename) + 1];
 
@@ -609,14 +638,12 @@ static struct module *do_module(const char *dirname,
 
 			if (is_higher_priority(newpath, (*i)->pathname,search,
 					       overrides)) {
-				new->next = (*i)->next;
-
-				release_file((*i)->data, (*i)->len);
-				free(*i);
-
+				del_module(i, *i);
+				
+				new->next = *i;
 				*i = new;
 			} else
-				free(new);
+				del_module(NULL, new);
 
 			return list;
 		}
@@ -679,7 +706,7 @@ static struct module *grab_basedir(const char *dirname,
 	return list;
 }
 
-static void parse_modules(struct module *list)
+static struct module *parse_modules(struct module *list)
 {
 	struct module *i;
 
@@ -687,28 +714,22 @@ static void parse_modules(struct module *list)
 		i->ops->load_symbols(i);
 		i->ops->fetch_tables(i);
 	}
-}
-
-/* Convert filename to the module name.  Works if filename == modname, too. */
-static void filename2modname(char *modname, const char *filename)
-{
-	const char *afterslash;
-	unsigned int i;
-
-	afterslash = strrchr(filename, '/');
-	if (!afterslash)
-		afterslash = filename;
-	else
-		afterslash++;
-
-	/* Convert to underscores, stop at first . */
-	for (i = 0; afterslash[i] && afterslash[i] != '.'; i++) {
-		if (afterslash[i] == '-')
-			modname[i] = '_';
-		else
-			modname[i] = afterslash[i];
+	
+	for (i = list; i; i = i->next)
+		i->ops->calculate_deps(i, verbose);
+	
+	/* Strip out modules with dependency loops. */
+ again:
+	for (i = list; i; i = i->next) {
+		if (has_dep_loop(i, NULL)) {
+			warn("Module %s ignored, due to loop\n",
+			     i->pathname + skipchars);
+			del_module(&list, i);
+			goto again;
+		}
 	}
-	modname[i] = '\0';
+	
+	return list;
 }
 
 /* Simply dump hash table. */
@@ -731,6 +752,35 @@ static void output_symbols(struct module *unused, FILE *out)
 	}
 }
 
+static void output_symbols_bin(struct module *unused, FILE *out)
+{
+	struct index_node *index;
+	unsigned int i;
+	char *line;
+
+	index = index_create();
+	
+	for (i = 0; i < SYMBOL_HASH_SIZE; i++) {
+		struct symbol *s;
+
+		for (s = symbolhash[i]; s; s = s->next) {
+			if (s->owner) {
+				char modname[strlen(s->owner->pathname)+1];
+				filename2modname(modname, s->owner->pathname);
+				asprintf(&line, "symbol:%s %s",
+					s->name, modname);
+				if (index_insert(index, line) && warn_dups)
+					warn("duplicate module syms:\n%s\n",
+						line);
+				free(line);
+			}
+		}
+	}
+	
+	index_write(index, out);
+	index_destroy(index);
+}
+
 static const char *next_string(const char *string, unsigned long *secsize)
 {
 	/* Skip non-zero chars */
@@ -749,6 +799,29 @@ static const char *next_string(const char *string, unsigned long *secsize)
 	return string;
 }
 
+/* Careful!  Don't munge - in [ ] as per Debian Bug#350915 */
+static char *underscores(char *string)
+{
+	unsigned int i;
+
+	if (!string)
+		return NULL;
+		
+	for (i = 0; string[i]; i++) {
+		switch (string[i]) {
+		case '-':
+			string[i] = '_';
+			break;
+		
+		case '[':
+			i += strcspn(&string[i], "]");
+			if (!string[i])
+				warn("Unmatched bracket in %s\n", string);
+		}
+	}
+	return string;
+}
+
 static void output_aliases(struct module *modules, FILE *out)
 {
 	struct module *i;
@@ -778,6 +851,52 @@ static void output_aliases(struct module *modules, FILE *out)
 	}
 }
 
+static void output_aliases_bin(struct module *modules, FILE *out)
+{
+	struct module *i;
+	const char *p;
+	char *line;
+	unsigned long size;
+	struct index_node *index;
+
+	index = index_create();
+	
+	for (i = modules; i; i = i->next) {
+		char modname[strlen(i->pathname)+1];
+
+		filename2modname(modname, i->pathname);
+
+		/* Grab from old-style .modalias section. */
+		for (p = i->ops->get_aliases(i, &size);
+		     p;
+		     p = next_string(p, &size)) {
+			asprintf(&line, "%s %s", p, modname);
+			underscores(line);
+			if (index_insert(index, line) && warn_dups)
+				warn("duplicate module alias:\n%s\n", line);
+			free(line);
+		}
+
+		/* Grab from new-style .modinfo section. */
+		for (p = i->ops->get_modinfo(i, &size);
+		     p;
+		     p = next_string(p, &size)) {
+			if (strncmp(p, "alias=", strlen("alias=")) == 0) {
+				asprintf(&line, "%s %s",
+					p + strlen("alias="), modname);
+				underscores(line);
+				if (index_insert(index, line) && warn_dups)
+					warn("duplicate module alias:\n%s\n",
+						line);
+				free(line);
+			}
+		}
+	}
+	
+	index_write(index, out);
+	index_destroy(index);
+}
+
 struct depfile {
 	char *name;
 	void (*func)(struct module *, FILE *);
@@ -785,6 +904,7 @@ struct depfile {
 
 static struct depfile depfiles[] = {
 	{ "modules.dep", output_deps }, /* This is what we check for '-A'. */
+	{ "modules.dep.bin", output_deps_bin },
 	{ "modules.pcimap", output_pci_table },
 	{ "modules.usbmap", output_usb_table },
 	{ "modules.ccwmap", output_ccw_table },
@@ -794,7 +914,9 @@ static struct depfile depfiles[] = {
 	{ "modules.ofmap", output_of_table },
 	{ "modules.seriomap", output_serio_table },
 	{ "modules.alias", output_aliases },
+	{ "modules.alias.bin", output_aliases_bin },
 	{ "modules.symbols", output_symbols },
+	{ "modules.symbols.bin", output_symbols_bin }
 };
 
 /* If we can't figure it out, it's safe to say "true". */
@@ -816,18 +938,22 @@ static int any_modules_newer(const char *dirname, time_t mtime)
 
 		sprintf(file, "%s/%s", dirname, dirent->d_name);
 		if (lstat(file, &st) != 0)
-			return 1;
+			goto ret_true;
 
 		if (smells_like_module(dirent->d_name)) {
 			if (st.st_mtime > mtime)
-				return 1;
+				goto ret_true;
 		} else if (S_ISDIR(st.st_mode)) {
 			if (any_modules_newer(file, mtime))
-				return 1;
+				goto ret_true;
 		}
 	}
 	closedir(dir);
 	return 0;
+
+ret_true:
+	closedir(dir);
+	return 1;
 }
 
 static int depfile_out_of_date(const char *dirname)
@@ -843,42 +969,48 @@ static int depfile_out_of_date(const char *dirname)
 	return any_modules_newer(dirname, st.st_mtime);
 }
 
-static int fgetc_wrapped(FILE *file, unsigned int *linenum)
-{
-	for (;;) {
-		int ch = fgetc(file);
-		if (ch != '\\')
-			return ch;
-		ch = fgetc(file);
-		if (ch != '\n')
-			return ch;
-		if (linenum)
-			(*linenum)++;
-	}
-}
-
 static char *getline_wrapped(FILE *file, unsigned int *linenum)
 {
-	int size = 1024;
+	int size = 256;
 	int i = 0;
 	char *buf = NOFAIL(malloc(size));
 	for(;;) {
-		int ch = fgetc_wrapped(file, linenum);
-		if (i == size) {
-			size *= 2;
-			buf = NOFAIL(realloc(buf, size));
-		}
-		if (ch < 0 && i == 0) {
-			free(buf);
-			return NULL;
-		}
-		if (ch < 0 || ch == '\n') {
+		int ch = getc_unlocked(file);
+		
+		switch(ch) {
+		case EOF:
+			if (i == 0) {
+				free(buf);
+				return NULL;
+			}
+			/* else fall through */
+			
+		case '\n':
 			if (linenum)
 				(*linenum)++;
+			if (i == size)
+				buf = NOFAIL(realloc(buf, size + 1));
 			buf[i] = '\0';
-			return NOFAIL(realloc(buf, i+1));
+			return buf;
+			
+		case '\\':
+			ch = getc_unlocked(file);
+			
+			if (ch == '\n') {
+				if (linenum)
+					(*linenum)++;
+				continue;
+			}
+			/* else fall through */
+		
+		default:
+			buf[i++] = ch;
+	
+			if (i == size) {
+				size *= 2;
+				buf = NOFAIL(realloc(buf, size));
+			}
 		}
-		buf[i++] = ch;
 	}
 }
 
@@ -948,8 +1080,10 @@ static int read_config_file(const char *filename,
 
 		cmd = strsep_skipspace(&ptr, "\t ");
 
-		if (cmd == NULL || cmd[0] == '#' || cmd[0] == '\0')
+		if (cmd == NULL || cmd[0] == '#' || cmd[0] == '\0') {
+			free(line);
 			continue;
+		}
 
 		if (strcmp(cmd, "search") == 0) {
 			char *search_path;
@@ -1053,27 +1187,27 @@ static void read_toplevel_config(const char *filename,
 
 	/* Try defaults. */
 	for (i = 0; i < ARRAY_SIZE(default_configs); i++) {
-		if (read_config(default_configs[i], basedir, search, overrides))
-			return;
+		read_config(default_configs[i], basedir, search, overrides);
 	}
 }
 
+/* Local to main, but not freed on exit.  Keep valgrind quiet. */
+struct module *list = NULL;
+struct module_search *search = NULL;
+struct module_overrides *overrides = NULL;
 
 int main(int argc, char *argv[])
 {
 	int opt, all = 0, maybe_all = 0, doing_stdout = 0;
 	char *basedir = "", *dirname, *version, *badopt = NULL,
 		*system_map = NULL;
-	struct module *list = NULL;
 	int i;
 	const char *config = NULL;
-	struct module_search *search = NULL;
-	struct module_overrides *overrides = NULL;
 
 	/* Don't print out any errors just yet, we might want to exec
            backwards compat version. */
 	opterr = 0;
-	while ((opt = getopt_long(argc, argv, "ab:ArehnqruvVF:C:", options, NULL))
+	while ((opt = getopt_long(argc, argv, "ab:ArehnqruvVF:C:w", options, NULL))
 	       != -1) {
 		switch (opt) {
 		case 'a':
@@ -1112,6 +1246,9 @@ int main(int argc, char *argv[])
 		case 'V':
 			printf("%s %s\n", PACKAGE, VERSION);
 			exit(0);
+		case 'w':
+			warn_dups = 1;
+			break;
 		default:
 			badopt = argv[optind-1];
 		}
@@ -1180,7 +1317,7 @@ int main(int argc, char *argv[])
 	} else {
 		list = grab_basedir(dirname,search,overrides);
 	}
-	parse_modules(list);
+	list = parse_modules(list);
 
 	for (i = 0; i < sizeof(depfiles)/sizeof(depfiles[0]); i++) {
 		FILE *out;
@@ -1196,8 +1333,11 @@ int main(int argc, char *argv[])
 			if (!out)
 				fatal("Could not open %s for writing: %s\n",
 					tmpname, strerror(errno));
-		} else
+		} else {
 			out = stdout;
+			if (ends_in(depname, ".bin"))
+				continue;
+		}
 		d->func(list, out);
 		if (!doing_stdout) {
 			fclose(out);
